---
toc: false
---

## Covariance operators

In many applications of fitting Gaussian processes at scale one uses a complex
or indirect method for working with large covariance matrices. For example, fast
multipole methods, hierarchical matrices, sparse approximations to precision
matrices, and other related strategies all provide ways to work with at least a
limited set of functionalities with a very large matrix that is not possible
using standard objects. 

In order to make this code as flexible and reusable as possible, it is possible
to provide these objects, presumably in the form of a custom-made `struct`, to
`gnll`. The `gnll` function will then use these arbitrary structs just as it
would a standard `Matrix` object, and only requires a select few operations,
which in Julia speak we would call methods. Importantly, if using SAA to
estimate the traces, all matrix-matrix operations are avoided, and so you really
only need a select few basic methods to implement.

The methods that an arbitrary covariance operator `M` needs to implement to be
compatible are:
* Updating with new parameters. This method is called `GPMaxlik.update!(M,
  locations, kernel, params)`.
* Solving a linear system with *some* symmetric factor of `M`, which is to say
  computing `W\x` and `W'\x` where $M == W*W'$. `W` does *not* need to be a
  triangular Cholesky factor. This method is called `GPMaxlik.sym_factor_solve(M,
  x; adj=true/false)`, where `adj` indicates whether it is the adjoint of `W` or
  not.
* A standard solve `M\x`, whose method is `Base.:\(M, x)`.

If you want to evaluate the likelihood, you will also need:
* A standard log-determinant (`LinearAlgebra.logdet(M)`).

With these methods defined, you can pass in your object `M` via `gnll(..., K=M,
...)` and everything will compose as you would expect.

## Derivatives of covariance operators

In order to compute the gradient and expected Fisher matrix, you will need first
derivatives of your object `M` with respect to kernel parameters. These can
again be entirely custom `struct`s of your own design. Your derivative operator,
which we'll call `Mj`, needs to implement the following methods:
* Updating with new parameters, just as above, except providing the derivative
  of the kernel with respect to the j-th parameter.
* A standard application, whose method is `Base.:*(Mj, x)`.

If you want to compute exact gradients and expected Fisher entries, you need to
additionally implement
* An exact trace function for `tr(M\Mj)`, which appears in the gradient. This
  method is `GPMaxlik.exact_grad_trace(M, Mj)`.
* An exact trace function for `tr((M\Mj)*(M\Mk))`, which appears in the expected
  Fisher matrix, and where `Mk` is a second derivative matrix operator. This
  method is `GPMaxlik.exact_efisher_term(M, Mj, Mk)`.

And that's it!

Here are a few caveats and additional pieces of information about this
functionality:
* If you many parameters and want an exact expected Fisher matrix as opposed to
  the (very high quality) stochastic estimator, consider providing more than two
derivative covariance operators `Mj`. `GPMaxlik` provides at least near-optimal
pathing in the assembly of the exact expected Fisher matrix to minimize calls to
`GPMaxlik.update!`.
[This](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/example/example_trustregion_prealloc.jl)
example file demonstrates the value of providing additional buffers as described
here, demonstrating also that this is useful even if you don't have any crazy
operator structs.
* `GPMaxlik` takes a conservative approach with user-provided operators--no
  copies are made. So you need to provide what you need for what you're asking
for. So if you want an exact expected Fisher matrix, you need to provide two
derivative objects.
* For a complete demo of creating these methods, see the implementation for the
  standard dense matrix objects
[here](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/src/covmatrix.jl).

