---
title: Documentation
toc: false
---

# `GPMaxlik.jl` Documentation

The package centers around one primary function called [gnll](./docs/gnll.md),
whose call encapsulates many of the options that the package offers. Some of
these are reasonably self-explanatory, such as `grad=false` meaning that you are
not interested in a gradient being returned, and so they won't be documented in
further detail than in the above list. But others are slightly more
sophisticated and will be covered in more detail here:

* For more information on the covariance operator interface, see
  [here](./docs/operators.md). For an example usage of providing operators, see
  [this](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/example/example_partial_prealloc.jl#L23)
  example file that demonstrates their usage in a complete application.  Note that
  this section is relevant even if you are just using the standard
  `CovarianceMatrix` object, because if you have a model with many parameters you
  can provide more than the minimum requirement of 2 derivative matrices (default
  `DerivativeCovarianceMatrix` objects) to reduce redundant work in assembling
  exact Fisher information matrices.

* For more information on the storage interface, see [here](./docs/storage.md).
  For an example usage of the storage interface, see
  [this](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/example/example_ipopt.jl#L23)
  example file that demonstrates their usage in a complete application.

* For more information on the options to automatically transition from
  stochastic trace estimation to exact trace estimation, see [here](./docs/storage.md).

* For more information on our provided trust region optimizer, see
  [here](./docs/trustregion.md).

# A briefer overview

The prototypical use of `GPMaxlik.jl` is to make a long function call that looks
like
```julia
  nll_call = gnll(data_locations, [data_1, data_2, ...], covariance_fun,
                  covariance_derivatives, parameters, saa=saa_input,
                  nll=true, grad=true, fish=true) 
```
which computes the log-likelihood, its gradient, and its expected Fisher matrix.
If `saa_input == nothing`, then it is computing the traces in the latter two
terms exactly. If not, it computes them using symmetrized Hutchinson trace
estimators (see the Geoga et al 2019 paper for details).  You would then access
the computed quantities with `nll_call.nll`, `nll_call.grad`, and
`nll_call.fish`.

If you want an exact likelihood, gradient, and **true hessian**, then you can
also use the function `gnll_forwarddiff`. This function does not give you the
option to provide any special structures or do much by way of operator
compartmentalization.  But what it does provide is special methods for the
gradient and Hessian that manually do the matrix calculus to compute those
quantities efficiently. So instead of using the generic `LinearAlgebra._chol!`
on a `Symmetric{Dual{T,V,N}, Matrix{Dual{T,V,N}}}`, this code will still use
LAPACK on `Matrix{Float64}` objects exclusively. For matrices of slightly
larger sizes, say above 2k, this can easily provide an order of magnitude
speedup in the Hessian in particular. The calling syntax is very easy:
```julia
  _nll(p) = GPMaxlik.gnll_forwarddiff(p, data_locations, data, kernel)
  ForwardDiff.gradient(_nll, p) # faster than naive forward-mode gradient
  ForwardDiff.hessian(_nll, p)  # WAY faster naive forward-mode Hessian
```
This won't necessarily open new cosmic possibilities for you, but if you want to
fit, say, 5k points to a model with 10 parameters, this could turn a runtime of an
hour to a runtime of, say, five minutes.

See the documentation for `gnll` for more details on all options, including the
more complicated ones such as storage and user-provided covariance operators.

