
import LinearAlgebra: BlasFloat

const DualHerm{T,V,N} = Hermitian{ForwardDiff.Dual{T,V,N}, Matrix{ForwardDiff.Dual{T,V,N}}}

# Getting the j-th FIRST derivative matrix from a Matrix{Dual}(...) where
# V<:BlasFloat, i.e. when you have called ForwardDiff.gradient and there is only
# one layer of partials.
function update_fderiv_buf!(Kj::Matrix{Float64}, K::DualHerm{T,V,N},
                            ix) where{T,V<:BlasFloat, N}
  @inbounds for _k in 1:size(K, 2)
    @inbounds for _j in 1:size(K, 1)
      Kj[_j,_k] = ForwardDiff.partials(K[_j,_k])[ix]
    end
  end
  nothing
end

# Getting the j-th FIRST derivative matrix from a Matrix{Dual}(...) where
# V<:Dual, i.e. when you have called ForwardDiff.hessian and there are two
# layers of partials.
function update_fderiv_buf!(Kj::Matrix{Float64}, K::DualHerm{T,V,N}, 
                            ix) where{T,V<:ForwardDiff.Dual, N}
  @inbounds for _k in 1:size(K, 2)
    @inbounds for _j in 1:size(K, 1)
      Kj[_j,_k] = ForwardDiff.partials(ForwardDiff.value(K[_j,_k]))[ix]
    end
  end
  nothing
end

# generic fallback.
function update_fderiv_buf!(Kj, K, ix)
  @inbounds for _k in 1:size(K, 2)
    @inbounds for _j in 1:size(K, 1)
      Kj[_j,_k] = ForwardDiff.partials(K[_j,_k])[ix]
    end
  end
  nothing
end

# Getting the (j,k)-th SECOND derivative matrix from a Matrix{Dual}(...) where
# V<:Dual, i.e. when you have called ForwardDiff.hessian and there are two
# layers of partials.
function update_sderiv_buf!(Kj::Matrix{Float64}, K::DualHerm{T,V,N}, 
                            ix1, ix2) where{T,V<:ForwardDiff.Dual, N}
  @inbounds for _k in 1:size(K, 2)
    @inbounds for _j in 1:size(K, 1)
      Kj[_j,_k] = ForwardDiff.partials(ForwardDiff.partials(K[_j,_k])[ix1])[ix2]
    end
  end
  nothing
end

# here "s*" means that * has already been solved, i.e. s* = Kf\*.
function gauss_nll_hessian_odiag(Kf, sKj, sKk, sKjk, data, sdata)
  # trace terms:
  trterm = -tr(sKk*sKj) + tr(sKjk)
  # quadratic form terms:
  qform  = -dot(data, sKk*(sKj*sdata))
  qform += dot(data, sKjk*sdata)
  qform -= dot(data, sKj*(sKk*sdata))
  0.5*(trterm - qform)
end

# here "s*" means that * has already been solved, i.e. s* = Kf\*.
function gauss_nll_hessian_diag(Kf, sKj, sKjj, data, sdata)
  # trace terms:
  trterm = -tr(sKj^2) + tr(sKjj)
  # quadratic form terms:
  qform  = -2*dot(data, sKj*(sKj*sdata))
  qform += dot(data, sKjj*sdata)
  0.5*(trterm - qform)
end

function assemble_kernel_matrix(kernel, pts, params)
  k11 = kernel(pts[1], pts[1], params)
  buf = Array{eltype(k11)}(undef, length(pts), length(pts))
  n   = length(pts)
  Threads.@threads for k in eachindex(pts)
    ptk = pts[k]
    buf[k,k] = kernel(ptk, ptk, params)
    for j in 1:(k-1)
      buf[j,k] = kernel(pts[j], ptk, params)
    end
  end
  Hermitian(buf, :U)
end

# Raw ikelihood:
function gnll_forwarddiff(params, locs, data, kernel)
  _K = assemble_kernel_matrix(kernel, locs, params) 
  K  = cholesky!(_K)
  0.5*(logdet(K) + sum(x->x^2, K.U'\data))
end

# Gradient:
function gnll_forwarddiff(p::AbstractVector{ForwardDiff.Dual{T,V,N}}, locs, 
                          data, kernel) where{T,V<:BlasFloat,N}
  # Assemble the kernel matrix, which is a Matrix{Dual}(...), and then extract
  # the value part (which for a gradient will be a Matrix{Float64} and can thus
  # use the LAPACK cholesky) and pre-allocate the derivative part:
  K   = assemble_kernel_matrix(kernel, locs, p) 
  Kf  = cholesky!(Hermitian(ForwardDiff.value.(K)))
  Kj  = Array{Float64}(undef, size(K))
  _Kj = copy(Kj)
  # Compute the solved data system once just as a general little optimization:
  presolved_data = Kf\data
  # Now loop over the derivative with respect to each element of p, update the
  # derivative matrix buffer, compute the quadratic form, do an in-place solve
  # to get Kf\Kj in order to compute tr(Kf\Kj), and then return the result:
  g = ntuple(N) do j
    update_fderiv_buf!(Kj, K, j)
    qform = dot(presolved_data, Hermitian(Kj), presolved_data)
    ldiv!(Kf, Kj)
    0.5*(tr(Kj)-qform)
  end
  # Return the value and all the computed partials (the gradient):
  ForwardDiff.Dual{T}(0.5*(logdet(Kf) + dot(data, presolved_data)), g...)
end

# Hessian:
getT(d::ForwardDiff.Dual{T,V,N}) where{T,V,N} = T
function gnll_forwarddiff(p::AbstractVector{ForwardDiff.Dual{T,V,N}}, 
                          locs, data, kernel) where{T,V<:ForwardDiff.Dual,N}
  # Assemble the kernel matrix, which is a Matrix{Dual}(...), and then extract
  # the value part (which for a gradient will be a Matrix{Float64} and can thus
  # use the LAPACK cholesky) and pre-allocate all derivative buffers.
  K   = assemble_kernel_matrix(kernel, locs, p)
  Kf  = cholesky!(Hermitian(map(x->ForwardDiff.value(ForwardDiff.value(x)), K)))
  Kj  = Array{Float64}(undef, size(K))
  Kk  = Array{Float64}(undef, size(K))
  Kjk = Array{Float64}(undef, size(K))
  # Compute the solved data system once just as a general little optimization:
  presolved_data = Kf\data
  # Allocate the gradient and Hessian buffers and start the brutal loop. 
  g = zeros(N)  
  h = zeros(N,N)
  for j in 1:N
    # Get the FIRST derivative matrix Kj:
    update_fderiv_buf!(Kj, K, j)
    grad_qf = dot(presolved_data, Hermitian(Kj), presolved_data)
    ldiv!(Kf, Kj)
    # Off-diagonal entries:
    for k in 1:(j-1)
      # Get the FIRST derivative matrix Kk:
      update_fderiv_buf!(Kk, K, k)
      ldiv!(Kf, Kk)
      # Get the SECOND derivative matrix K_{j,k}:
      update_sderiv_buf!(Kjk, K, j, k)
      ldiv!(Kf, Kjk)
      # fill in values:
      h[j,k] = gauss_nll_hessian_odiag(Kf, Kj, Kk, Kjk, data, presolved_data)
      h[k,j] = h[j,k]
    end
    # Manually fill in the diagonal element:
    update_sderiv_buf!(Kjk, K, j, j)
    ldiv!(Kf, Kjk)
    h[j,j] = gauss_nll_hessian_diag(Kf, Kj, Kjk, data, presolved_data)
    # do a gradient term while we're here:
    g[j] = 0.5*(tr(Kj) - grad_qf)
  end
  _T   = getT(ForwardDiff.partials(p[1])[1])
  _val = ForwardDiff.Dual{_T}(0.5*(logdet(Kf) + dot(data, presolved_data)), g...)
  _pts = map(j->ForwardDiff.Dual{_T}(g[j], h[:,j]...), 1:N)
  ForwardDiff.Dual{T}(_val, _pts...)
end

