---
title: '\texttt{GPMaxlik.jl}: A suite of tools for efficient maximum likelihood optimization of parametric Gaussian process models'
bibliography: references.bib
tags:
  - Julia
  - statistics
  - Gaussian processes
  - stochastic trace estimation
authors:
  - name: Christopher J. Geoga
    affiliation: "1,3"
    email: "christopher.geoga@rutgers.edu" 
  - name: Paul G. Beckman
    affiliation: "2"
affiliations:
  - name: Rutgers University, Dept. of Statistics
    index: 1
  - name: Courant Insitute, New York University
    index: 2
  - name: Argonne National Laboratory, Dept. of Mathematics and Computer Science
    index: 3
#date: 12 March, 2022
author: Christopher J. Geoga\thanks{Rutgers University, Dept. of Statistics} \thanks{Argonne National Laboratory, Dept. of Mathematics and Computer Science} \quad and \quad Paul G. Beckman\thanks{Courant Insitute, New York University}
---

\newcommand{\bx}{\mathbf{x}} \newcommand{\by}{\mathbf{y}}
\newcommand{\bu}{\mathbf{u}} \newcommand{\bth}{\mathbf{\theta}}
\newcommand{\bS}{\mathbf{\Sigma}}

# Summary

`GPMaxlik.jl` provides a flexible interface to compartmentalize efficient
operations with large dense matrices from efficient evaluation of the Gaussian
log-likelihood and its derivatives. Users can provide arbitrary objects
representing covariance matrices capable of a minimal set of operations to the
primary `gnll` function and obtain the likelihood and its derivatives
efficiently and with minimal matrix operations, which at large data sizes are
necessarily the bottleneck of any parameter estimation routine. For ease of use,
it also provides a careful implementation of the likelihood and its derivatives
in the case where covariance matrices are not approximated. Finally, it offers
several advanced tools to accelerate the computation of derivatives of the
likelihood. These tools include:

* Joint computation of the likelihood and derivative information to exploit
  redundant intermediate quantities.

* Providing the expected Fisher information matrix as an effective proxy for the
  true Hessian of the log-likelihood, requiring only first derivatives of the
  kernel with respect to kernel parameters.

* The option to use *symmetrized* stochastic trace estimation for the trace
  terms in the derivatives of the Gaussian likelihood, which fully removes all
  matrix-matrix operations from both the gradient and expected Fisher
  information matrix, and computes both quantities in a single pass over the
  derivatives of the covariance function with respect to parameters.

* The option to provide a variable number of pre-allocated derivative matrices
  (again possibly arbitrary structures provided by the user) and near-optimal
  pathing in the creation of the exact expected Fisher information matrix to
  reduce the number of times the derivative matrices need to be updated.

* The option to cache intermediate results that will accelerate subsequent
  computations of stochastic traces, for example in constructing the stochastic
  expected Fisher information matrix after computing the gradient at the same
  values. This provides meaningful speed gains when interfacing with software
  that requires separate functions for gradients and Hessians, such as Ipopt
  [@wachter2006].

* A special method for the likelihood with exact linear algebra that has
  manually defined methods for use with `ForwardDiff.jl`, both for the gradient
  and true Hessian, that is careful to perform linear algebra only on
  `Matrix{Float64}` objects, thus exploiting BLAS/LAPACK libraries.

# A brief introduction to Gaussian process computations

Gaussian processes are a popular tool in a broad range of statistical
applications. Using them at scale, however, is difficult, as in general one has
to deal with very large dense matrices. Worse still, the most common workflow of
specifying a parametric covariance function $K(\bx, \bx')$ such that
$\text{Cov}(Z(\bx), Z(\bx')) = K(\bx, \bx')$ for a process $Z$ (see [@stein1999]
for more information) and then estimating the parameters using data involves
working with large matrices inside a challenging nonconvex optimization problem.
For a mean-zero process, the likelihood to be minimized, after removing constant
terms, is given by $$ -2 \ell(\bth) = \log |\bS(\bth)| + \by^T \bS(\bth)^{-1}
\by, $$ where $\bS_{j,k} = K_{\bth}(\bx_j, \bx_k)$.  Due to the challenging
nature of the optimization problem, first- and even second-derivative
information is very useful for successful and efficient parameter estimation,
further adding to the already difficult collection of matrix operations
necessary to perform effective estimation. The gradient of $\ell$ has a closed
form and has terms given by $$ -[2 \nabla \ell(\bth)]_j = \text{tr}
\left(\bS(\bth)^{-1} \left[ \frac{\partial}{\partial \theta_j} \bS(\bth) \right]
\right) 
  - \by^T \bS(\bth)^{-1} \left[ \frac{\partial}{\partial \theta_j} \bS(\bth)
       \right] \bS(\bth)^{-1} \by. $$ The expected Fisher information matrix,
which can serve as a valuable proxy for true Hessian information, is given by 
$$
\mathcal{I}_{j,k} = \frac{1}{2} \text{tr}\left( \bS(\bth)^{-1} \left[
\frac{\partial}{\partial \theta_j} \bS(\bth) \right] \bS(\bth)^{-1} \left[
\frac{\partial}{\partial \theta_k} \bS(\bth) \right] \right). $$ 
Both of these quantities have several challenging aspects. Most importantly,
they contain matrix-matrix operations. The quadratic form in the gradient can of
course be computed in such a way that avoids any matrix-matrix operations, but
the traces are not so easily handled. @stein2013 introduced the application of
Hutchinson trace estimation [@hutchinson1990] to this problem and discussed a 
further enhancement when symmetric factors for matrices are
available. First applied in @geoga2019, these enhancements were demonstrated to
be of significant value, showing that the approximation $$ \text{tr} \left(
\mathbf{A} \mathbf{B} \right) \approx \frac{1}{N} \sum_{j=1}^N \bu_j^T
\mathbf{L}^T \mathbf{A} \mathbf{L} \bu_j, $$ where $\mathbf{B} = \mathbf{L}
\mathbf{L}^T$, can be used to obtain estimators for the traces in both the
gradient and expected Fisher information matrix above that are accurate enough
for numerical optimization. Additional work is required to use this
"symmetrization" trick in the expected Fisher information matrix as it is
offered in \texttt{GPMaxlik.jl}, but we refer the curious reader to
@geoga2019.

# Statement of need

Many methods have been introduced to combat the poor scaling of dense linear
algebra for large covariance matrices. The majority of these methods take
advantage of the fact that common kernels are smooth away from the origin, which
leads to matrix blocks that are very well approximated by low rank matrices (an
observation that dates back to at least the fast multipole method
[@Greengard87]). Such methods include hierarchically off-diagonal low-rank
(HODLR) matrix structures ([@Ambikasaran2016, @geoga2019], strongly-admissible
hierarchical matrices [@Litvinenko2017], closely related ``skeletonization"
methods ([@Minden2016]), and many others. A related but separate field of
methods instead approximates the inverse of the covariance matrix with a sparse
substitute (see, for example, @Rue2005 and @katzfuss2017 and references within
for more information).

Although basic code for evaluating the Gaussian log-likelihood and its gradient
is straightforward, an implementation that is not sufficiently careful runs the
risk of introducing significant redundant work, or potentially even requiring
matrix-matrix operations that may not be implemented or possible in good
complexity for more exotic matrix approximations. At present, most projects
designed to apply fast matrix algorithms to Gaussian processes involve an at
least partial re-implementation of these quantities. Not only does this fragment
the software ecosystem and make it harder for practitioners to try multiple
methods, it presents the serious risk of practitioners being hindered by
mistakes or inefficiencies in the implementation of the likelihood or its
derivatives.

`GPMaxlik.jl` attempts to solve this problem by offering a flexible interface
that accepts any matrix structure and offers a number of relevant optional
features, several of which are not offered elsewhere. It is our hope that this
will not only reduce the work barrier to applying new matrix algorithms to the
problem of Gaussian process computations, but will also make it much easier to
perform more effective optimization using our infrastructure for efficient
stochastic gradients and expected Fisher information matrices.

# Applications in published and ongoing work

`GPMaxlik.jl` is a mature software package that has been used in several
published works and in ongoing projects dating back to 2019.

* The basic trust region-based optimizer (see @nocedal2006 for more information)
  was used in @geoga2019.

* @geoga2021 used the stochastic trace optimization to change the expected
  runtime of an optimization problem from nearly a week to hours, as the model
  has 25 parameters and involves matrices of dimension approximately 17,000. The
  appendix to [@geoga2021] provides back-of-the-envelope math, but squeezing out
  all matrix-matrix operations from the derivative information in the likelihood
  was all but a requirement to make the computations practically feasible.

* @beckman2022 fits a 192-parameter model to over 100,000 measurements. This
  setting in particular is one where a single computation of expected Fisher
  information matrix without stochastic trace estimation takes nearly a week of
  computer time. The stochastic analog requires less than 12 hours.

All three of these projects use different matrix structures and yet the same
interface and optimizer, demonstrating that the problem of efficiently working
with large covariance matrices can be compartmentalized from the problem of
efficient likelihood and derivative evaluations. This package provides a common
software layer for these methods which would otherwise be burdensome to
re-implement as needed with every project.

# Acknowledgements

Argonne National Laboratory’s work was supported by the US Department of Energy, Office of Science, Office of Advanced Scientific Computing Research under contract DE-AC02-06CH11357

The authors are grateful Michael L. Stein (Rutgers University, Dept. of
Statistics) and Mihai Anitescu (University of Chicago, Dept. of Statistics and
Argonne National Laboratory, Dept. of Mathematics and Computer Science) for
their mentorship over years of projects, including all of the projects in which
this code has been used, and in the earlier projects that inspired the creation
of such a compatibilty layer for rapid development of future methods.

\noindent\fbox{\parbox{\columnwidth}{\footnotesize
    The submitted manuscript has been created by UChicago Argonne, LLC, Operator of Argonne National Laboratory (``Argonne''). Argonne, a U.S. Department of Energy Office of Science laboratory, is operated under Contract No. DE-AC02-06CH11357. The U.S. Government retains for itself, and others acting on its behalf, a paid-up nonexclusive, irrevocable worldwide license in said article to reproduce, prepare derivative works, distribute copies to the public, and perform publicly and display publicly, by or on behalf of the Government. The Department of Energy will provide public access to these results of federally sponsored research in accordance with the DOE Public Access Plan (http://energy.gov/downloads/doe-public-access-plan).}
}

# References


